#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of Rok.
#
# Copyright © 2020 Arrikto Inc.  All Rights Reserved.

"""Script to generate a ServiceAccountToken for an external client."""

import os
import sys
import enum
import logging

from kubernetes import client, config
from kubernetes.client import models

from rok_tasks import frontend, question
from rok_tasks.frontend import Choice
from rok_tasks.frontend.cli import RokCLI

log = logging.getLogger(__name__)
fr = frontend.get_frontend()

TITLE = "Kubernetes ServiceAccount-Role Binder"
DESCRIPTION = "Bind a ServiceAccount with a Role on Kubernetes"


# Helpers
######################################################


def load_kube_config():
    # Load K8s configuration, if present
    if "KUBERNETES_SERVICE_HOST" in os.environ:
        config.load_incluster_config()
    else:
        try:
            config.load_kube_config()
        except IOError as e:
            raise RuntimeError("No K8s configuration was found: %s" % e)


def create_role_binding(api_client, rb):
    namespace = rb.metadata.namespace
    k8s_client = client.RbacAuthorizationV1Api(api_client)
    return k8s_client.create_namespaced_role_binding(namespace, rb)


class Role(enum.Enum):
    """Enumaration of supported roles."""

    VIEW = "view"
    EDIT = "edit"
    ADMIN = "admin"

    @property
    def desc(self):
        if self is self.VIEW:
            return "Viewer"
        elif self is self.EDIT:
            return "Editor"
        elif self is self.ADMIN:
            return "Administrator"
        else:
            raise RuntimeError("Invalid role option: %s" % self)


# Questions
######################################################


class ServiceAccountName(question.LineInputQuestion):
    """Question for Service Account Name."""

    name = "question/service_account_name"
    msg = "What is the name of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-name"],
        "metavar": "SERVICE_ACCOUNT_NAME",
        "help": "Use ServiceAccount with name %(metavar)s"
    }


class ServiceAccountNamespace(question.LineInputQuestion):
    """Question for Service Account Namespace."""

    name = "question/service_account_namespace"
    msg = "What is the namespace of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-namespace"],
        "metavar": "SERVICE_ACCOUNT_NAMESPACE",
        "help": "Use ServiceAccount with namespace %(metavar)s"
    }


class Role(question.MenuQuestion):
    """Question for Role."""

    name = "question/role"
    msg = "What is the role you wish to grant to the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--role"],
        "metavar": "ROLE",
        "help": "Bind ServiceAccount with role %(metavar)s"
    }
    choices = [Choice(id=r.value, desc=r.desc, long_desc="") for r in Role]


class RoleBindingNamespace(question.LineInputQuestion):
    """Question for RoleBinding Namespace."""

    def __init__(self, *args, **kwargs):
        super(RoleBindingNamespace, self).__init__(*args, **kwargs)
        sa_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
        if os.path.exists(sa_path):
            with open(sa_path) as f:
                self.default = f.readline()

    name = "question/role_binding_namespace"
    msg = "What is the namespace you want to give access to?"
    required = True
    argument_opts = {
        "name": ["--role-binding-namespace"],
        "help": "Namespace to give access to"
    }


# CLI
######################################################


class ServiceAccountRoleBinder(RokCLI):
    """Create a ServiceAccount and ServiceAccountToken."""

    IDENTIFIABLE_LABELS = {"arrikto.com/service-account-token-gen": ""}
    DEFAULT_LOGFILE = "service_account_role_binder.log"
    DEFAULT_FRONTEND = "readline"

    def initialize_parser(self):
        parser = super(ServiceAccountRoleBinder, self).initialize_parser()
        questions = [ServiceAccountName(), ServiceAccountNamespace(), Role(),
                     RoleBindingNamespace()]
        self.questions = questions
        question.add_questions_args(questions, parser)
        return parser

    def _init_clients(self):
        load_kube_config()
        self.api_client = client.ApiClient()

    def _ask_questions(self):
        # TODO: Ask all questions
        for q in self.questions:
            self.qctx.add_question(q)
            q.ask()
        return

    def make_role_binding(self):
        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        role = self.qctx.questions[Role.name].get_answer()
        rb_namespace = (self.qctx.questions[RoleBindingNamespace.name]
                        .get_answer())

        kubeflow_role = "kubeflow-%s" % role
        rb_name = "%s-%s-%s" % (sa_namespace, sa_name, role)
        user_name = "system:serviceaccount:%s:%s" % (sa_namespace, sa_name)

        metadata = models.V1ObjectMeta(annotations={"user": user_name,
                                                    "role": role},
                                       name=rb_name, namespace=rb_namespace,
                                       labels=self.IDENTIFIABLE_LABELS)
        role_ref = models.V1RoleRef(api_group="rbac.authorization.k8s.io",
                                    kind="ClusterRole", name=kubeflow_role)
        subject = models.V1Subject(api_group="rbac.authorization.k8s.io",
                                   kind="User", name=user_name)
        return models.V1RoleBinding(api_version="rbac.authorization.k8s.io/v1",
                                    kind="RoleBinding", metadata=metadata,
                                    role_ref=role_ref, subjects=[subject])

    def main(self):
        self._init_clients()
        self._ask_questions()

        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        role = self.qctx.questions[Role.name].get_answer()
        rb_namespace = (self.qctx.questions[RoleBindingNamespace.name]
                        .get_answer())

        with fr.Progress(maxval=2, title=TITLE) as p:

            p.info("Generating RoleBinding...")
            rb = self.make_role_binding()
            p.inc()

            p.info("Applying RoleBinding...")
            create_role_binding(self.api_client, rb)
            p.inc()

            p.success("Successfully created RoleBinding")

        msg = (u"Successfully granted `%s' access in namespace `%s' to the"
               u" ServiceAccount `%s/%s'.\n"
               % (role, rb_namespace, sa_namespace, sa_name))

        with fr.Message(msg, title=TITLE, lvl=frontend.DEBUG,
                        ok_button=True) as m:
            m.show()


def main():
    cli = ServiceAccountRoleBinder()
    return cli.run()


if __name__ == "__main__":
    sys.exit(main())
